#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_utils


cdef extern from "azure_c_shared_utility/xio.h":

    ctypedef void* CONCRETE_IO_HANDLE
    ctypedef void* XIO_HANDLE

    cdef enum IO_SEND_RESULT_TAG:
        IO_SEND_OK,
        IO_SEND_ERROR,
        IO_SEND_CANCELLED

    cdef enum IO_OPEN_RESULT_TAG:
        IO_OPEN_OK,
        IO_OPEN_ERROR,
        IO_OPEN_CANCELLED

    ctypedef void (*ON_BYTES_RECEIVED)(void* context, const unsigned char* buffer, size_t size)
    ctypedef void (*ON_SEND_COMPLETE)(void* context, IO_SEND_RESULT_TAG send_result)
    ctypedef void (*ON_IO_OPEN_COMPLETE)(void* context, IO_OPEN_RESULT_TAG open_result)
    ctypedef void (*ON_IO_CLOSE_COMPLETE)(void* context)
    ctypedef void (*ON_IO_ERROR)(void* context)

    ctypedef c_utils.OPTIONHANDLER_HANDLE (*IO_RETRIEVEOPTIONS)(CONCRETE_IO_HANDLE concrete_io)
    ctypedef CONCRETE_IO_HANDLE (*IO_CREATE)(void* io_create_parameters)
    ctypedef void (*IO_DESTROY)(CONCRETE_IO_HANDLE concrete_io)
    ctypedef int (*IO_OPEN)(CONCRETE_IO_HANDLE concrete_io, ON_IO_OPEN_COMPLETE on_io_open_complete, void* on_io_open_complete_context, ON_BYTES_RECEIVED on_bytes_received, void* on_bytes_received_context, ON_IO_ERROR on_io_error, void* on_io_error_context)
    ctypedef int (*IO_CLOSE)(CONCRETE_IO_HANDLE concrete_io, ON_IO_CLOSE_COMPLETE on_io_close_complete, void* callback_context)
    ctypedef int (*IO_SEND)(CONCRETE_IO_HANDLE concrete_io, const void* buffer, size_t size, ON_SEND_COMPLETE on_send_complete, void* callback_context)
    ctypedef void (*IO_DOWORK)(CONCRETE_IO_HANDLE concrete_io)
    ctypedef int (*IO_SETOPTION)(CONCRETE_IO_HANDLE concrete_io, const char* optionName, const void* value)

    ctypedef struct IO_INTERFACE_DESCRIPTION_TAG:
        IO_RETRIEVEOPTIONS concrete_io_retrieveoptions
        IO_CREATE concrete_io_create
        IO_DESTROY concrete_io_destroy
        IO_OPEN concrete_io_open
        IO_CLOSE concrete_io_close
        IO_SEND concrete_io_send
        IO_DOWORK concrete_io_dowork
        IO_SETOPTION concrete_io_setoption

    ctypedef IO_INTERFACE_DESCRIPTION_TAG IO_INTERFACE_DESCRIPTION

    XIO_HANDLE xio_create(const IO_INTERFACE_DESCRIPTION* io_interface_description, const void* io_create_parameters)
    void xio_destroy(XIO_HANDLE xio)
    int xio_open(XIO_HANDLE xio, ON_IO_OPEN_COMPLETE on_io_open_complete, void* on_io_open_complete_context, ON_BYTES_RECEIVED on_bytes_received, void* on_bytes_received_context, ON_IO_ERROR on_io_error, void* on_io_error_context)
    int xio_close(XIO_HANDLE xio, ON_IO_CLOSE_COMPLETE on_io_close_complete, void* callback_context)
    int xio_send(XIO_HANDLE xio, const void* buffer, size_t size, ON_SEND_COMPLETE on_send_complete, void* callback_context)
    void xio_dowork(XIO_HANDLE xio)
    int xio_setoption(XIO_HANDLE xio, const char* optionName, const void* value)
    c_utils.OPTIONHANDLER_HANDLE xio_retrieveoptions(XIO_HANDLE xio)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
from enum import Enum
import time
import logging

# C imports
from libc cimport stdint
from libc.stdlib cimport malloc, free
from libc.string cimport memset

cimport c_cbs
cimport c_utils
cimport c_strings
cimport c_session


_logger = logging.getLogger(__name__)


cpdef create_sas_token(const char* key, const char* scope, const char* keyname, size_t expiry):
    cdef c_strings.STRING_HANDLE str_value
    #cdef const char* c_string
    str_value = c_utils.SASToken_CreateString(key, scope, keyname, expiry)

    if <void*>str_value == NULL:
        raise ValueError("Failed to create SAS token.")
    if c_utils.SASToken_Validate(str_value) != True:
        raise ValueError("Generated invalid SAS token")
    return c_strings.STRING_c_str(str_value)
    #c_strings.STRING_delete(str_value)
    #return c_string


cdef class CBSTokenAuth(object):

    cdef const char* audience
    cdef const char* token_type
    cdef const char* token
    cdef stdint.uint64_t expires_at
    cdef stdint.uint64_t _refresh_window
    cdef c_cbs.CBS_HANDLE _cbs_handle
    cdef c_cbs.AUTH_STATUS state
    cdef stdint.uint64_t auth_timeout
    cdef stdint.uint64_t _token_put_time
    cdef unsigned int token_status_code
    cdef const char* token_status_description
    cdef const char* connection_id
    cdef cSession _session

    def __cinit__(self, const char* audience, const char* token_type, const char* token, stdint.uint64_t expires_at, cSession session, stdint.uint64_t timeout, const char* connection_id):
        self.state = AUTH_STATUS_IDLE
        self.audience = audience
        self.token_type = token_type
        self.token = token
        self.expires_at = expires_at
        self.auth_timeout = timeout
        self.connection_id = connection_id
        self._token_put_time = 0
        current_time = int(time.time())
        remaining_time = expires_at - current_time
        self._refresh_window = int(float(remaining_time) * 0.1)
        self._cbs_handle = c_cbs.cbs_create(<c_session.SESSION_HANDLE>session._c_value)
        self._session = session
        if <void*>self._cbs_handle == NULL:
            raise MemoryError("Unable to create CBS Handle.")
        if c_cbs.cbs_open_async(self._cbs_handle, <c_cbs.ON_CBS_OPEN_COMPLETE>on_cbs_open_complete, <void*>self, <c_cbs.ON_CBS_ERROR>on_cbs_error, <void*>self) != 0:
            raise ValueError("Unable to open CBS link.")

    def __dealloc__(self):
        _logger.debug("Deallocating CBSTokenAuth")
        self.destroy()

    cpdef destroy(self):
        if <void*>self._cbs_handle is not NULL:
            _logger.debug("Destroying CBSTokenAuth for connection %r", self.connection_id)
            c_cbs.cbs_destroy(self._cbs_handle)
            self._cbs_handle = <c_cbs.CBS_HANDLE>NULL
            self._session = None

    cpdef close(self):
        if c_cbs.cbs_close(self._cbs_handle) != 0:
            self._value_error("Unable to close CBS link.")

    cpdef set_trace(self, bint trace_on):
        if c_cbs.cbs_set_trace(self._cbs_handle, trace_on) != 0:
            raise ValueError("Unable to set debug trace.")

    cpdef authenticate(self):
        if self.state == AUTH_STATUS_IN_PROGRESS:
            return
        current_time = int(time.time())
        if current_time >= self.expires_at:
            raise ValueError("Token has expired")
        self._token_put_time = current_time
        if c_cbs.cbs_put_token_async(
                self._cbs_handle,
                self.token_type,
                self.audience,
                self.token,
                <c_cbs.ON_CBS_OPERATION_COMPLETE>on_cbs_put_token_complete,
                <void*>self) != 0:
            raise ValueError("Put-Token request failed.")
        else:
            self.state = AUTH_STATUS_IN_PROGRESS

    cpdef get_status(self):
        self._update_status()
        return self.state

    cpdef get_failure_info(self):
        return self.token_status_code, self.token_status_description

    cpdef refresh(self, const char* refresh_token, stdint.uint64_t expires_at):
        self._update_status()
        if self.state == AUTH_STATUS_REFRESH_REQUIRED:
            self.token = refresh_token
            self.authenticate()
            self.expires_at = expires_at

    cpdef _update_status(self):
        error_code = 0
        is_refresh_required = False
        if self.state == AUTH_STATUS_OK or self.state == AUTH_STATUS_REFRESH_REQUIRED:
            is_expired, is_refresh_required = self._check_expiration_and_refresh_status()
            if is_expired:
                self.state = AUTH_STATUS_EXPIRED
            elif is_refresh_required:
                self.state = AUTH_STATUS_REFRESH_REQUIRED
        elif self.state == AUTH_STATUS_IN_PROGRESS:
            put_timeout = self._check_put_timeout_status()
            if put_timeout:
                self.state = AUTH_STATUS_TIMEOUT

    cpdef _check_put_timeout_status(self):
        seconds_since_epoc = int(time.time())
        if self.auth_timeout > 0:
            return (seconds_since_epoc - self._token_put_time) >= self.auth_timeout
        else:
            return False

    cpdef _check_expiration_and_refresh_status(self):
        seconds_since_epoc = int(time.time())
        is_expired = seconds_since_epoc >= self.expires_at
        is_refresh_required = (self.expires_at - seconds_since_epoc) <= self._refresh_window
        return is_expired, is_refresh_required

    cpdef _cbs_open_complete(self, result):
        self.on_cbs_open_complete(result)

    cpdef on_cbs_open_complete(self, result):
        _logger.info("CBS for connection %r completed opening with status: %r", self.connection_id, result)
        if result == c_cbs.CBS_OPEN_COMPLETE_RESULT_TAG.CBS_OPEN_ERROR:
            self.state = AUTH_STATUS_FAILURE

    cpdef _cbs_error(self):
        self.on_cbs_error()

    cpdef on_cbs_error(self):
        _logger.info("CBS error occured on connection %r.", self.connection_id)

    cpdef _cbs_put_token_compelete(self, c_cbs.CBS_OPERATION_RESULT_TAG result, unsigned int status_code, const char* status_description):
        if result == CBS_OPERATION_RESULT_OK:
            self.state = AUTH_STATUS_OK
        else:
            self.state = AUTH_STATUS_ERROR
        self.token_status_code = status_code
        self.token_status_description = status_description
        self.on_cbs_put_token_complete(result, status_code, status_description)

    cpdef on_cbs_put_token_complete(self, c_cbs.CBS_OPERATION_RESULT_TAG result, unsigned int status_code, const char* status_description):
        _logger.info("Token put complete with result: %r, status: %r, description: %r, connection: %r", result, status_code, status_description, self.connection_id)


#### Callbacks

cdef void on_cbs_open_complete(void *context, c_cbs.CBS_OPEN_COMPLETE_RESULT_TAG open_complete_result):
    if <void*>context != NULL:
        context_pyobj = <PyObject*>context
        if context_pyobj.ob_refcnt == 0: # context is being garbage collected, skip the callback
            _logger.warning("Can't call on_cbs_open_complete during garbage collection.")
            return
        context_obj = <object>context
        context_obj._cbs_open_complete(open_complete_result)


cdef void on_cbs_error(void* context):
    if <void*>context != NULL:
        context_pyobj = <PyObject*>context
        if context_pyobj.ob_refcnt == 0: # context is being garbage collected, skip the callback
            _logger.warning("Can't call on_cbs_error during garbage collection.")
            return
        context_obj = <object>context
        context_obj._cbs_error()


cdef void on_cbs_put_token_complete(void* context, c_cbs.CBS_OPERATION_RESULT_TAG complete_result, unsigned int status_code, const char* status_description):
    cdef unsigned int verified_status_code
    cdef const char* verified_description
    if <void*>context != NULL:
        context_pyobj = <PyObject*>context
        if context_pyobj.ob_refcnt == 0: # context is being garbage collected, skip the callback
            _logger.warning("Can't call on_cbs_put_token_complete during garbage collection.")
            return
        context_obj = <object>context
        try:
            if <void*>status_code != NULL:
                verified_status_code = status_code
            else:
                verified_status_code = 0
            if <void*>status_description != NULL:
                verified_description = status_description
            else:
                verified_description = b"CBS Session closed."
            context_obj._cbs_put_token_compelete(complete_result, verified_status_code, verified_description)
        except KeyboardInterrupt:
            context_obj._cbs_put_token_compelete(
                CBS_OPERATION_RESULT_INSTANCE_CLOSED, 1, b"Client shutdown with keyboard interrupt.")



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
cimport c_message_receiver
cimport c_message
cimport c_link
cimport c_amqpvalue



_logger = logging.getLogger(__name__)


cpdef create_message_receiver(cLink link, callback_context):
    receiver = cMessageReceiver()
    receiver.create(link, on_message_receiver_state_changed, <void*>callback_context)
    return receiver


cdef class cMessageReceiver(StructBase):

    cdef c_message_receiver.MESSAGE_RECEIVER_HANDLE _c_value
    cdef const char* _link_name
    cdef cLink _link

    def __cinit__(self):
        pass

    def __dealloc__(self):
        _logger.debug("Deallocating cMessageReceiver")
        self.destroy()

    cdef _validate(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cdef create(self, cLink link, c_message_receiver.ON_MESSAGE_RECEIVER_STATE_CHANGED on_message_sender_state_changed, void* context):
        self.destroy()
        self._link = link
        self._c_value = c_message_receiver.messagereceiver_create(<c_link.LINK_HANDLE>link._c_value, on_message_sender_state_changed, context)
        self._validate()
        if c_message_receiver.messagereceiver_get_link_name(self._c_value, &self._link_name)!= 0:
            self._value_error("Unable to retrieve message receiver link name.")

    cpdef open(self, callback_context):
        if c_message_receiver.messagereceiver_open(self._c_value, <c_message_receiver.ON_MESSAGE_RECEIVED>on_message_received, <void*>callback_context) != 0:
            self._value_error()

    cpdef close(self):
        if c_message_receiver.messagereceiver_close(self._c_value) != 0:
            self._value_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destroying cMessageReceiver")
            c_message_receiver.messagereceiver_destroy(self._c_value)
            self._c_value = <c_message_receiver.MESSAGE_RECEIVER_HANDLE>NULL
            self._link = None

    cpdef last_received_message_number(self):
        cdef c_amqp_definitions.delivery_number message_number
        if c_message_receiver.messagereceiver_get_received_message_id(self._c_value, &message_number) != 0:
            self._value_error("Unable to retrieve last received message number.")
        return message_number

    cpdef settle_accepted_message(self, c_amqp_definitions.delivery_number message_number):
        cdef c_amqpvalue.AMQP_VALUE delivery_state
        delivery_state = c_message.messaging_delivery_accepted()
        if c_message_receiver.messagereceiver_send_message_disposition(self._c_value, self._link_name, message_number, delivery_state) != 0:
            raise RuntimeError("Unable to send message dispostition 'accepted' for message number {}".format(message_number))
        c_amqpvalue.amqpvalue_destroy(delivery_state)

    cpdef settle_released_message(self, c_amqp_definitions.delivery_number message_number):
        cdef c_amqpvalue.AMQP_VALUE delivery_state
        delivery_state = c_message.messaging_delivery_released()
        if c_message_receiver.messagereceiver_send_message_disposition(self._c_value, self._link_name, message_number, delivery_state) != 0:
            raise RuntimeError("Unable to send message dispostition 'released' for message number {}".format(message_number))
        c_amqpvalue.amqpvalue_destroy(delivery_state)

    cpdef settle_rejected_message(self, c_amqp_definitions.delivery_number message_number, const char* error_condition, const char* error_description):
        cdef c_amqpvalue.AMQP_VALUE delivery_state
        delivery_state = c_message.messaging_delivery_rejected(error_condition, error_description)
        if c_message_receiver.messagereceiver_send_message_disposition(self._c_value, self._link_name, message_number, delivery_state) != 0:
            raise RuntimeError("Unable to send message dispostition 'rejected' for message number {}".format(message_number))
        c_amqpvalue.amqpvalue_destroy(delivery_state)

    cpdef settle_modified_message(self, c_amqp_definitions.delivery_number message_number, bint delivery_failed, bint undeliverable_here, AMQPValue annotations):
        cdef c_amqpvalue.AMQP_VALUE delivery_state
        cdef c_amqp_definitions.fields delivery_fields
        if annotations is not None:
            delivery_fields = <c_amqp_definitions.fields>annotations._c_value
        else:
            delivery_fields = <c_amqp_definitions.fields>NULL
        delivery_state = c_message.messaging_delivery_modified(delivery_failed, undeliverable_here, delivery_fields)
        if c_message_receiver.messagereceiver_send_message_disposition(self._c_value, self._link_name, message_number, delivery_state) != 0:
            raise RuntimeError("Unable to send message dispostition 'delivery-modified' for message number {}".format(message_number))
        c_amqpvalue.amqpvalue_destroy(delivery_state)

    cdef wrap(self, cMessageReceiver value):
        self.destroy()
        self._link = value._link
        self._c_value = value._c_value
        self._validate()

    cpdef set_trace(self, bint value):
        c_message_receiver.messagereceiver_set_trace(self._c_value, value)


#### Callbacks (context is a MessageReceiver instance)

cdef void on_message_receiver_state_changed(void* context, c_message_receiver.MESSAGE_RECEIVER_STATE_TAG new_state, c_message_receiver.MESSAGE_RECEIVER_STATE_TAG previous_state):
    if context != NULL:
        context_pyobj = <PyObject*>context
        if context_pyobj.ob_refcnt == 0: # context is being garbage collected, skip the callback
            _logger.warning("Can't call on_state_changed during garbage collection, please be sure to close or use a context manager")
            return
        context_obj = <object>context
        try:
            context_obj._state_changed(previous_state, new_state)
        except AttributeError:
            _logger.info("Unknown MessageReceiver state changed: %r to %r", previous_state, new_state)


cdef c_amqpvalue.AMQP_VALUE on_message_received(void* context, c_message.MESSAGE_HANDLE message):
    cdef c_message.MESSAGE_HANDLE cloned
    cloned = c_message.message_clone(message)
    wrapped_message = message_factory(cloned)

    if context != NULL:
        context_pyobj = <PyObject*>context
        if context_pyobj.ob_refcnt == 0: # context is being garbage collected, skip the callback
            _logger.warning("Can't call _message_received during garbage collection")
            return <c_amqpvalue.AMQP_VALUE>NULL
        context_obj = <object>context
        if hasattr(context_obj, '_message_received'):
            context_obj._message_received(wrapped_message)
    return <c_amqpvalue.AMQP_VALUE>NULL



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
from libc cimport stdint

cimport c_amqpvalue
cimport c_amqp_definitions


_logger = logging.getLogger(__name__)


cpdef create_header():
    new_header = cHeader()
    return new_header


cdef class cHeader(StructBase):

    cdef c_amqp_definitions.HEADER_HANDLE _c_value

    def __cinit__(self):
        self._c_value = c_amqp_definitions.header_create()
        self._validate()

    def __dealloc__(self):
        _logger.debug("Deallocating cHeader")
        self.destroy()

    cdef _validate(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        try:
            if <void*>self._c_value is not NULL:
                _logger.debug("Destroying cHeader")
                c_amqp_definitions.header_destroy(self._c_value)
        except KeyboardInterrupt:
            pass
        finally:
            self._c_value = <c_amqp_definitions.HEADER_HANDLE>NULL

    cdef wrap(self, c_amqp_definitions.HEADER_HANDLE value):
        self.destroy()
        self._c_value = value
        self._validate()

    cpdef clone(self):
        cdef c_amqp_definitions.HEADER_HANDLE value
        value = c_amqp_definitions.header_clone(self._c_value)
        if <void*>value == NULL:
            self._value_error()
        new_obj = cHeader()
        new_obj.wrap(value)
        return new_obj

    @property
    def delivery_count(self):
        cdef stdint.uint32_t _value
        if c_amqp_definitions.header_get_delivery_count(self._c_value, &_value) == 0:
            if <void*>_value == NULL:
                return None
            return _value
        else:
            return None

    @delivery_count.setter
    def delivery_count(self, stdint.uint32_t value):
        if c_amqp_definitions.header_set_delivery_count(
            self._c_value, value) != 0:
                self._value_error("Couldn't set 'delivery_count'.")

    @property
    def time_to_live(self):
        cdef c_amqp_definitions.milliseconds _value
        if c_amqp_definitions.header_get_ttl(self._c_value, &_value) == 0:
            if <void*>_value == NULL:
                return None
            return _value
        else:
            return None

    @time_to_live.setter
    def time_to_live(self, c_amqp_definitions.milliseconds value):
        if c_amqp_definitions.header_set_ttl(
            self._c_value, value) != 0:
                self._value_error("Couldn't set 'time_to_live'.")

    @property
    def durable(self):
        cdef bint _value
        if c_amqp_definitions.header_get_durable(self._c_value, &_value) == 0:
            if <void*>_value == NULL:
                return None
            return _value
        else:
            return None

    @durable.setter
    def durable(self, bint value):
        if c_amqp_definitions.header_set_durable(
            self._c_value, value) != 0:
                self._value_error("Couldn't set 'durable'.")

    @property
    def first_acquirer(self):
        cdef bint _value
        if c_amqp_definitions.header_get_first_acquirer(self._c_value, &_value) == 0:
            if <void*>_value == NULL:
                return None
            return _value
        else:
            return None

    @first_acquirer.setter
    def first_acquirer(self, bint value):
        if c_amqp_definitions.header_set_first_acquirer(
            self._c_value, value) != 0:
                self._value_error("Couldn't set 'first_acquirer'.")

    @property
    def priority(self):
        cdef stdint.uint8_t _value
        if c_amqp_definitions.header_get_priority(self._c_value, &_value) == 0:
            if <void*>_value == NULL:
                return None
            return _value
        else:
            return None

    @priority.setter
    def priority(self, stdint.uint8_t value):
        if c_amqp_definitions.header_set_priority(
            self._c_value, value) != 0:
                self._value_error("Couldn't set 'priority'.")


#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging
import copy

# C imports
from libc cimport stdint

from cpython.ref cimport PyObject
cimport c_message_sender
cimport c_link
cimport c_async_operation
cimport c_amqpvalue


_logger = logging.getLogger(__name__)


cpdef create_message_sender(cLink link, callback_context):
    sender = cMessageSender()
    sender.create(link, on_message_sender_state_changed, <void*>callback_context)
    return sender


cdef create_message_sender_with_callback(cLink link,c_message_sender.ON_MESSAGE_SENDER_STATE_CHANGED callback, void* callback_context):
    sender = cMessageSender()
    sender.create(link, callback, callback_context)
    return sender


cdef class cMessageSender(StructBase):

    cdef c_message_sender.MESSAGE_SENDER_HANDLE _c_value
    cdef cLink _link

    def __cinit__(self):
        pass

    def __dealloc__(self):
        _logger.debug("Deallocating cMessageSender")
        self.destroy()

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.destroy()

    cpdef open(self):
        if c_message_sender.messagesender_open(self._c_value) != 0:
            self._value_error()

    cpdef close(self):
        if c_message_sender.messagesender_close(self._c_value) != 0:
            self._value_error()

    cdef _create(self):
        if <void*>self._c_value is NULL:
            self._memory_error()

    cpdef destroy(self):
        if <void*>self._c_value is not NULL:
            _logger.debug("Destroying cMessageSender")
            c_message_sender.messagesender_destroy(self._c_value)
            self._c_value = <c_message_sender.MESSAGE_SENDER_HANDLE>NULL
            self._link = None

    cdef wrap(self, cMessageSender value):
        self.destroy()
        self._link = value._link
        self._c_value = value._c_value
        self._create()

    cdef create(self, cLink link, c_message_sender.ON_MESSAGE_SENDER_STATE_CHANGED on_message_sender_state_changed, void* context):
        self.destroy()
        self._link = link
        self._c_value = c_message_sender.messagesender_create(<c_link.LINK_HANDLE>link._c_value, on_message_sender_state_changed, context)
        self._create()

    cpdef send(self, cMessage message, c_amqp_definitions.tickcounter_ms_t timeout, callback_context):
        operation = c_message_sender.messagesender_send_async(self._c_value, <c_message.MESSAGE_HANDLE>message._c_value, on_message_send_complete, <void*>callback_context, timeout)
        if <void*>operation is NULL:
            _logger.info("Send operation result is NULL")
            return False
        return True

    cpdef set_trace(self, bint value):
        c_message_sender.messagesender_set_trace(self._c_value, value)


#### Callbacks (context is a MessageSender instance)


cdef void on_message_send_complete(void* context, c_message_sender.MESSAGE_SEND_RESULT_TAG send_result, c_amqpvalue.AMQP_VALUE delivery_state):
    cdef c_amqpvalue.AMQP_VALUE send_data
    if <void*>delivery_state == NULL:
        wrapped = None
    else:
        send_data = c_amqpvalue.amqpvalue_clone(delivery_state)
        wrapped = copy.deepcopy(value_factory(send_data).value)
    if context != NULL:
        context_pyobj = <PyObject*>context
        if context_pyobj.ob_refcnt == 0: # context is being garbage collected, skip the callback
            _logger.warning("Can't call _on_message_sent during garbage collection")
            return
        context_obj = <object>context
        if hasattr(context_obj, "_on_message_sent"):
            context_obj._on_message_sent(context_obj, send_result, delivery_state=wrapped)


cdef void on_message_sender_state_changed(void* context, c_message_sender.MESSAGE_SENDER_STATE_TAG new_state, c_message_sender.MESSAGE_SENDER_STATE_TAG previous_state):
    if context != NULL:
        context_pyobj = <PyObject*>context
        if context_pyobj.ob_refcnt == 0: # context is being garbage collected, skip the callback
            _logger.warning("Can't call on_state_changed during garbage collection, please be sure to close or use a context manager")
            return
        context_obj = <object>context
        try:
            context_obj._state_changed(previous_state, new_state)
        except AttributeError:
            _logger.info("Unknown MessageSender state changed: %r to %r", previous_state, new_state)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# pylint: disable=super-init-not-called,arguments-differ

import six
from uamqp import c_uamqp, compat, utils


class AMQPType(object):
    """Base type for specific AMQP encoded type definitions.

    :ivar value: The Python value of the AMQP type.
    :ivar c_data: The C AMQP encoded object.
    """

    def __init__(self, value):
        self._c_type = self._c_wrapper(value)

    @property
    def value(self):
        return self._c_type.value

    @property
    def c_data(self):
        return self._c_type

    def _c_wrapper(self, value):
        raise NotImplementedError()


class AMQPSymbol(AMQPType):
    """An AMQP symbol object.

    :ivar value: The Python value of the AMQP type.
    :vartype value: bytes
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: c_uamqp.SymbolValue
    :param value: The value to encode as an AMQP symbol.
    :type value: bytes or str
    :param encoding: The encoding to be used if a str is provided.
     The default is 'UTF-8'.
    :type encoding: str
    """

    def __init__(self, value, encoding='UTF-8'):
        self._c_type = self._c_wrapper(value, encoding)

    def _c_wrapper(self, value, encoding='UTF-8'):
        value = value.encode(encoding) if isinstance(value, six.text_type) else value
        return c_uamqp.symbol_value(value)


class AMQPChar(AMQPType):
    """An AMQP char object.

    :ivar value: The Python value of the AMQP type.
    :vartype value: bytes
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: c_uamqp.CharValue
    :param value: The value to encode as an AMQP char.
    :type value: bytes or str
    :param encoding: The encoding to be used if a str is provided.
     The default is 'UTF-8'.
    :type encoding: str
    """

    def __init__(self, value, encoding='UTF-8'):
        self._c_type = self._c_wrapper(value, encoding)

    def _c_wrapper(self, value, encoding='UTF-8'):
        if len(value) > 1:
            raise ValueError("Value must be a single character.")
        value = value.encode(encoding) if isinstance(value, six.text_type) else value
        return c_uamqp.char_value(value)


class AMQPLong(AMQPType):
    """An AMQP long object.

    :ivar value: The Python value of the AMQP type.
    :vartype value: int
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: uamqp.c_uamqp.LongValue
    :param value: The value to encode as an AMQP ulong.
    :type value: int
    :raises: ValueError if value is not within allowed range.
    """

    def _c_wrapper(self, value):
        try:
            return c_uamqp.long_value(compat.long(value))
        except TypeError:
            raise ValueError("Value must be an integer")
        except OverflowError:
            raise ValueError("Value {} is too large for a Long value.".format(value))


class AMQPuLong(AMQPType):
    """An AMQP unsigned long object.

    :ivar value: The Python value of the AMQP uLong.
    :vartype value: int
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: uamqp.c_uamqp.ULongValue
    :param value: The value to encode as an AMQP unsigned Long.
    :type value: list
    :raises: ValueError if value is not within allowed range.
    """

    def _c_wrapper(self, value):
        try:
            return c_uamqp.ulong_value(compat.long(value))
        except TypeError:
            raise ValueError("Value must be an integer")
        except OverflowError:
            raise ValueError("Value {} is too large for an unsigned Long value.".format(value))


class AMQPByte(AMQPType):
    """An AMQP byte object.

    :ivar value: The Python value of the AMQP type.
    :vartype value: int
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: uamqp.c_uamqp.ByteValue
    :param value: The value to encode as an AMQP ulong.
    :type value: int
    :raises: ValueError if value is not within allowed range.
    """

    def _c_wrapper(self, value):
        try:
            return c_uamqp.byte_value(int(value))
        except TypeError:
            raise ValueError("Value must be an integer")
        except OverflowError:
            raise ValueError("Value {} is too large for a Byte value.".format(value))


class AMQPuByte(AMQPType):
    """An AMQP unsigned byte object.

    :ivar value: The Python value of the AMQP uByte.
    :vartype value: int
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: uamqp.c_uamqp.UByteValue
    :param value: The value to encode as an AMQP unsigned Byte.
    :type value: list
    :raises: ValueError if value is not within allowed range.
    """

    def _c_wrapper(self, value):
        try:
            return c_uamqp.ubyte_value(int(value))
        except TypeError:
            raise ValueError("Value must be an integer")
        except OverflowError:
            raise ValueError("Value {} is too large for an unsigned Byte value.".format(value))


class AMQPInt(AMQPType):
    """An AMQP int object.

    :ivar value: The Python value of the AMQP type.
    :vartype value: int
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: uamqp.c_uamqp.IntValue
    :param value: The value to encode as an AMQP int.
    :type value: int
    :raises: ValueError if value is not within allowed range.
    """

    def _c_wrapper(self, value):
        try:
            return c_uamqp.int_value(int(value))
        except TypeError:
            raise ValueError("Value must be an integer")
        except OverflowError:
            raise ValueError("Value {} is too large for an Int value.".format(value))


class AMQPuInt(AMQPType):
    """An AMQP unsigned int object.

    :ivar value: The Python value of the AMQP uInt.
    :vartype value: int
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: uamqp.c_uamqp.UIntValue
    :param value: The value to encode as an AMQP unsigned int.
    :type value: list
    :raises: ValueError if value is not within allowed range.
    """

    def _c_wrapper(self, value):
        try:
            return c_uamqp.uint_value(int(value))
        except TypeError:
            raise ValueError("Value must be an integer")
        except OverflowError:
            raise ValueError("Value {} is too large for an unsigned int value.".format(value))


class AMQPArray(AMQPType):
    """An AMQP Array object. All the values in the array
    must be of the same type.

    :ivar value: The Python values of the AMQP array.
    :vartype value: list
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: uamqp.c_uamqp.ArrayValue
    :param value: The value to encode as an AMQP array.
    :type value: int
    :raises: ValueError if all values are not the same type.
    """

    def _c_wrapper(self, value_array):
        if value_array:
            value_type = type(value_array[0])
            if not all(isinstance(x, value_type) for x in value_array):
                raise ValueError("All Array values must be the same type.")

        c_array = c_uamqp.array_value()
        for value in value_array:
            c_array.append(utils.data_factory(value))
        return c_array


class AMQPDescribed(AMQPType):
    """An AMQP Described object. All the values in the array
    must be of the same type.

    :ivar value: The Python values of the AMQP array.
    :vartype value: list
    :ivar c_data: The C AMQP encoded object.
    :vartype c_data: uamqp.c_uamqp.ArrayValue
    :param value: The value to encode as an AMQP array.
    :type value: int
    :raises: ValueError if all values are not the same type.
    """

    def __init__(self, descriptor, described):
        self._c_type = self._c_wrapper(descriptor, described)

    def _c_wrapper(self, descriptor, described):
        descriptor = utils.data_factory(descriptor)
        described = utils.data_factory(described)
        return c_uamqp.described_value(descriptor, described)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
from libc cimport stdint

cimport c_amqpvalue
cimport c_amqp_definitions
cimport c_utils


_logger = logging.getLogger(__name__)


cpdef create_data(char* binary_data):
    cdef c_amqpvalue.AMQP_VALUE body_data
    cdef c_amqpvalue.amqp_binary _binary
    length = len(binary_data)
    _binary.length = length
    _binary.bytes = <void*>binary_data
    body_data = c_amqp_definitions.amqpvalue_create_data(_binary)
    if <void*>body_data == NULL:
        raise ValueError("Unable to create payload data")
    return value_factory(body_data)


cpdef create_sequence(AMQPValue sequence_data):
    body_data = c_amqp_definitions.amqpvalue_create_amqp_sequence(<c_amqpvalue.AMQP_VALUE>sequence_data._c_value)
    if <void*>body_data == NULL:
        raise ValueError("Unable to create payload data")
    return value_factory(body_data)



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

from libc cimport stdint

cimport c_xio


cdef extern from "azure_c_shared_utility/tlsio.h":

    ctypedef struct TLSIO_CONFIG_TAG:
        const char* hostname
        int port
        const c_xio.IO_INTERFACE_DESCRIPTION* underlying_io_interface
        void* underlying_io_parameters

    ctypedef TLSIO_CONFIG_TAG TLSIO_CONFIG


cdef extern from "azure_c_shared_utility/http_proxy_io.h":

    ctypedef struct HTTP_PROXY_IO_CONFIG_TAG:
        const char* hostname
        int port
        const char* proxy_hostname
        int proxy_port
        const char* username
        const char* password

    ctypedef HTTP_PROXY_IO_CONFIG_TAG HTTP_PROXY_IO_CONFIG

    const c_xio.IO_INTERFACE_DESCRIPTION* http_proxy_io_get_interface_description()


#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

# Python imports
import logging

# C imports
cimport c_tlsio
cimport c_utils


_logger = logging.getLogger(__name__)


cdef class StructBase(object):
    """Base class for wrapped C structs."""

    def _memory_error(self):
        message = "Failed to allocate memory to create {}"
        raise MemoryError(message.format(self.__class__.__name__))

    def _value_error(self, error_message=None, error_code=None):
        message = "Operation failed."
        if error_message:
            message += "\nError: {}".format(error_message)
        if error_code:
            message += "\nErrorCode: {}".format(error_code)
        raise ValueError(message)

    def _null_error(self, error_message=None):
        message = "NULL error occurred in {}.".format(self.__class__.__name__)
        if error_message:
            message += "\nError: {}".format(error_message)
        raise ValueError(message)


cdef class TickCounter(object):

    cdef c_utils.TICK_COUNTER_HANDLE _c_value

    def __cinit__(self):
        self._c_value = c_utils.tickcounter_create()
        if <void*>self._c_value == NULL:
            raise MemoryError("Failed to create tick counter.")

    def __dealloc__(self):
        self.destroy()

    cpdef destroy(self):
        if <void*>self._c_value != NULL:
            c_utils.tickcounter_destroy(self._c_value)
            self._c_value = <c_utils.TICK_COUNTER_HANDLE>NULL

    cpdef get_current_ms(self):
        cdef c_utils.tickcounter_ms_t current_ms
        if c_utils.tickcounter_get_current_ms(self._c_value, &current_ms) != 0:
            raise ValueError("Failed to get current ms time.")
        return current_ms
